/*
 * zoom.c Copyright Jan Panteltje 2005 and later.
 *
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

	{
	int l, pc = 0;
	int average, average_r, average_g, average_b;
	int h_limit;

	if(debug_flag)
		{
		fprintf(stderr,\
		"zoom: zoom_factor=%d image_bytes_per_color=%d display_width=%d display_height=%d\n\
		ptri=%p ptro=%p zoom_x_size=%d view_width=%d\n",\
		zoom_factor, image_bytes_per_color, display_width, display_height,\
		ptri, ptro, zoom_x_size, view_width);		
		}

	average = average_r =  average_g = average_b = 0; // -Wall
	l = 0; // -Wall

	/* if no zoom this may be a little faster, but not much, because we have to test h_limit */
	if(zoom_factor == 1)
		{
		if(enlarge_factor == 1)
			{
			for(i = 0; i < display_height; i++)
				{
				pa = i * view_width * zoom_factor * image_bytes_per_color;

				for(j = 0; j < display_width; j++)
					{
					pb = image_bytes_per_color * j * zoom_factor;

					h_limit = pa + (image_bytes_per_color * zoom_x_size);

//e = (3 * i * view_width * zoom_factor) + (3 * j * zoom_factor) + (3 * k * view_width) + (3 * l);
					e = pa + pb;

					if(image_bytes_per_color == 1)
						{
						if(e > h_limit) 
							{
							a = 0;
							ptri++;
							}
						else
							{
							b = *ptri & 0xff;
							ptri++;

							a = gamma_table[b];
							}

						*ptro = (a << 16) | (a << 8) |  a;		
						ptro++;
						}
					else if(image_bytes_per_color == 3)
						{
						if(e > h_limit)
							{
							r = 0;
							g = 0;
							b = 0;	
							ptri += 3;
							}
						else
							{
							a = *ptri & 0xff;
							b = gamma_table[a];
							ptri++;

							a = *ptri & 0xff;
							g = gamma_table[a];
							ptri++;

							a = *ptri & 0xff;
							r = gamma_table[a];
							ptri++;
							}

						*ptro = (b << 16) | (g << 8) | r;
						ptro++;
						}
					} /* end for j */
				} /* end for i */
			} /* end if enlarge_factor == 1 */
		else /* enlarge_factor > 1 */
			{
			for(i = 0; i < display_height; i++)
				{
				/* pointer to line source image */
				pa = (i / enlarge_factor) * (view_width * zoom_factor * image_bytes_per_color);

				for(j = 0; j < display_width; j++)
					{
					/* pointer to byte in line in source image */
					pb = (j / enlarge_factor) * (image_bytes_per_color * zoom_factor);

					h_limit = pa + (image_bytes_per_color * zoom_x_size);

//e = (3 * i * view_width * zoom_factor) + (3 * j * zoom_factor) + (3 * k * view_width) + (3 * l);
					e = pa + pb;
//fprintf(stderr, "i=%d j=%d pa=%d pb=%d e=%d\n", i, j, pa, pb, e);
//if(i == 4) exit(1);

					if(image_bytes_per_color == 1)
						{
						if(e > h_limit) 
							{
							a = 0;
							}
						else
							{
							b = ptri[e] & 0xff;

							a = gamma_table[b];
							}

						*ptro = (a << 16) | (a << 8) |  a;		
						ptro++;
						}
					else if(image_bytes_per_color == 3)
						{
						if(e > h_limit)
							{
							r = 0;
							g = 0;
							b = 0;	
							}
						else
							{
							a = ptri[e] & 0xff;
							b = gamma_table[a];

							a = ptri[e + 1] & 0xff;
							g = gamma_table[a];

							a = ptri[e + 2] & 0xff;
							r = gamma_table[a];
							}

						*ptro = (b << 16) | (g << 8) | r;
						ptro++;
						}
					} /* end for j */
				} /* end for i */
			} /* end if expand > 1 */
		} /* end if zoom_factor == 1 */
	else if(zoom_factor > 1)/* use zoom */
		{ 
		d = zoom_factor * zoom_factor;
		for(i = 0; i < display_height; i++)
			{
			pa = i * view_width * zoom_factor * image_bytes_per_color;

			for(j = 0; j < display_width; j++)
				{
				if(image_bytes_per_color == 1)
					{
					pb = j * zoom_factor;

					/* average over an area zoom_factor pixels vertical * zoom_factor pixels horizontal */ 		
					average = 0;
					for(k = 0; k < zoom_factor; k++)
						{
						pc = k * view_width;

						h_limit = pa + pc + zoom_x_size;

						for(l = 0; l < zoom_factor; l++)
							{
							e = pa + pb + pc + l;

							if(e > h_limit)
								{
								average = 0;
								}
							else
								{
								if(e > view_size) break;	

								average += ptri[e] & 0xff;
								}
							}
						}

					average /= d;

					a = gamma_table[average];

					*ptro = (a << 16) | (a << 8) |  a;		
					ptro++;
					} /* end if image_bytes_per_color == 1 */
				else if(image_bytes_per_color == 3)
					{
					pb = 3 * j * zoom_factor;

					/* average over an area zoom_factor pixels vertical * zoom_factor pixels horizontal */ 		
					average_r = 0;
					average_g = 0;
					average_b = 0;

					for(k = 0; k < zoom_factor; k++)
						{
						pc = 3 * k * view_width;

						h_limit = pa + pc + (3 * zoom_x_size);

						for(l = 0; l < zoom_factor; l++)
							{
							e = pa + pb + pc + (3 * l);

//		e = (3 * i * view_width * zoom_factor) + (3 * j * zoom_factor) + (3 * k * view_width) + (3 * l);  

							if(e > h_limit)
								{
								average_r = 0;
								average_g = 0;
								average_b = 0;
								}
							else
								{
								if(e > view_size * 3) break;	

								average_r += (ptri[e] & 0xff);
								average_g += (ptri[e + 1] & 0xff);
								average_b += (ptri[e + 2] & 0xff);
								}
							}
						}

					average_r /= d;
					b = gamma_table[average_r];

					average_g /= d;
					g = gamma_table[average_g];

					average_b /= d;
					r = gamma_table[average_b]; 

					*ptro = (b << 16) | (g << 8) | r;
					ptro++;
					} /* end if image_bytes_per_color == 3 */

				} /* end for j */
			} /* end for i */
		} /* end if zoom > 1 */
	else if(zoom_factor < 1)
		{
		d = zoom_factor * zoom_factor;
		for(i = 0; i < display_height; i++)
			{
			pa = i * view_width * zoom_factor * image_bytes_per_color;

			for(j = 0; j < display_width; j++)
				{
				if(image_bytes_per_color == 1)
					{
					pb = j * zoom_factor;

					a = ptri[e] & 0xff;

					a = gamma_table[average];

					*ptro = (a << 16) | (a << 8) |  a;		
					ptro++;
					} /* end if image_bytes_per_color == 1 */
				else if(image_bytes_per_color == 3)
					{
					pb = 3 * j * zoom_factor;

					e = pa + pb + pc + (3 * l);

					if(e > view_size * 3) break;	

					r += (ptri[e] & 0xff);
					g += (ptri[e + 1] & 0xff);
					b += (ptri[e + 2] & 0xff);


					b = gamma_table[average_r];

					g = gamma_table[average_g];

					r = gamma_table[average_b]; 

					*ptro = (b << 16) | (g << 8) | r;
					ptro++;
					} /* end if image_bytes_per_color == 3 */

				} /* end for j */
			} /* end for i */
		
		} /* end if zoom < 1 */

	if(debug_flag)
		{
		fprintf(stderr, "zoom complete\n");
		}

	}
